/*
 *   <one line to give the program's name and a brief idea of what it does.>
 *   Copyright (C) 2011 Shahin Boluki & Mohammadhosein Malmir & Amirahmad Barekatein  <sh.boluki@gmail.com>  & <mhmalmir@gmail.com> & <amirahmad.barekatein@gmail.com>
 * 
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 * 
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */



#include "bhv_mark.h"
#include "bhv_mark_action.h"
#include "bhv_set_play.h"
#include <rcsc/geom/vector_2d.h>
#include <rcsc/player/soccer_action.h>
#include <vector>
#include <rcsc/player/player_agent.h>
#include <rcsc/common/server_param.h>
#include "strategy.h"
#include <rcsc/math_util.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/composite_region_2d.h>
#include <rcsc/action/basic_actions.h>
#include <rcsc/action/bhv_go_to_point_look_ball.h>
#include <rcsc/action/body_go_to_point.h>

Vector2D Bhv_Mark_Action::last_m_pos;
Vector2D Bhv_Mark_Action::last_m_dead;
int      Bhv_Mark_Action::unums[12][12];

  

bool Bhv_Mark_Action::markActionExecute(PlayerAgent* agent, Vector2D& targetPoint, double& dashPower)
{
  const WorldModel & wm = agent->world();
  MarkRoute status ;
  status.target_pos_.invalidate() ;
  status.dash_power = Strategy::get_normal_dash_power(wm);


  if( (target_opp_unum < 1 || target_opp_unum > 11) )
    return false;
  if(!wm.theirPlayer(target_opp_unum))
    return false; 
  

///TODO: MarkPass
    if(wm.ball().posValid() && wm.ball().pos().x<-20.0 /*&& wm.ball().pos().absY()>11*/ && wm.theirPlayer(target_opp_unum)->pos().x<-20)
    {
	status = markPass(agent);
	  status.dash_power = ServerParam::i().maxDashPower();
    }
///TODO: ELSE !
    else
    { 
      	status = mark(agent);
        status.dash_power = Strategy::get_normal_dash_power(wm);
    }

 if (status.target_pos_.isValid() &&
     status.target_pos_.absX()<=ServerParam::i().pitchHalfLength() &&
     status.target_pos_.absY()<=ServerParam::i().pitchHalfWidth() &&
     ( ( status.target_pos_.dist2(Strategy::i().getPosition(wm.self().unum()))<= 144.0 ) ) )
 {
    targetPoint =  status.target_pos_ ;
    dashPower = status.dash_power ;
    return true;

 }

   targetPoint=Strategy::i().getPosition(wm.self().unum());
   dashPower = status.dash_power ;
   return false;

}

/*!*****************************************************************************************************************/

Bhv_Mark_Action::MarkRoute Bhv_Mark_Action::markPass(PlayerAgent* agent)
{  
   MarkRoute Mark_p;
  //Vector2D Mark_p;
  if( target_opp_unum < 1 || target_opp_unum > 11 )
  {
    Mark_p.mark_type_ = MARK_PASS;
    Mark_p.target_pos_.invalidate();
    return Mark_p;
  }
    
  const WorldModel & wm = agent->world();
  Vector2D opp_pos = wm.theirPlayer( target_opp_unum )->pos() + wm.theirPlayer( target_opp_unum )->vel();
  Vector2D opp_next_pos = wm.theirPlayer( target_opp_unum )->pos() + wm.theirPlayer( target_opp_unum )->vel();
  Vector2D ball_pos = wm.ball().pos() + wm.ball().vel();
  Vector2D vec_mark = Vector2D::INVALIDATED;
  Vector2D intersectPoint[2];
 
  
  double markCircleR =fabs(2 * wm.theirPlayer(target_opp_unum)->playerTypePtr()->kickableMargin());
  
  
 
  Circle2D markCircle( opp_next_pos , markCircleR );
  
  Segment2D mark_seg( opp_next_pos , ball_pos );
  Segment2D line_goal( opp_next_pos , Vector2D( -ServerParam::i().pitchHalfLength() , 0.0 ) );
 
  if( markCircle.intersection( mark_seg , &intersectPoint[0] , &intersectPoint[1] ) == 1 && (wm.self().pos() + wm.self().vel()).dist2(opp_next_pos)<400) 
  {
    vec_mark = ( intersectPoint[0].isValid() ) ? intersectPoint[0] : intersectPoint[1];

      
	Mark_p.mark_type_ = MARK_PASS;
	if((wm.self().pos() + wm.self().vel()).x - (wm.theirPlayer(target_opp_unum)->pos()+wm.theirPlayer(target_opp_unum)->vel()).x < 12)
	return Mark_p;
  }
  
  Mark_p.mark_type_ = MARK_PASS;
  Mark_p.target_pos_ = Vector2D::INVALIDATED;
  return Mark_p;
}


/*!************************************************************************************************************/
Bhv_Mark_Action::MarkRoute Bhv_Mark_Action::mark(PlayerAgent* agent)
{
 
  MarkRoute Mark_p;
  Mark_p.mark_type_ = MARK;
  Mark_p.target_pos_.invalidate();
  //Vector2D Mark_p;
  if( target_opp_unum < 1 || target_opp_unum > 11 )
  {
    Mark_p.mark_type_ = MARK;
    Mark_p.target_pos_.invalidate();
    return Mark_p;
  }
    
  const WorldModel & wm = agent->world();
  Vector2D home_pos=Strategy::i().getPosition(wm.self().unum());
  Vector2D opp_pos = wm.theirPlayer( target_opp_unum )->pos() + wm.theirPlayer( target_opp_unum )->vel();
  Vector2D opp_next_pos = wm.theirPlayer( target_opp_unum )->pos() + wm.theirPlayer( target_opp_unum )->vel();
  Vector2D ball_pos = wm.ball().pos() + wm.ball().vel();
  Vector2D vec_mark = Vector2D::INVALIDATED;
  Vector2D intersectPoint[2];

  return Mark_p;
}

     
     /*!**************************************************************************************************************/
     
     Bhv_Mark_Action::MarkRoute Bhv_Mark_Action::getBestMarkAction(PlayerAgent* agent)
     {
       const WorldModel & wm = agent->world();
       
       const double mark_type_score_weight = 1.0;
       
       const double mark_pass_teammate_six_x_weight = 0.0;
       const double mark_teammate_six_x_weight = 0.0;
       const double guard_teammate_six_x_weight = 0.0;
       const double positioning_teammate_six_x_weight = 0.0;
       const double teammate_six_x = wm.ourPlayer( 6 )->pos().x;
       
       const double mark_pass_ball_pos_x_weight = 0.0;
       const double mark_ball_pos_x_weight = 0.0;
       const double guard_ball_pos_x_weight = 0.0;
       const double positioning_ball_pos_x_weight = 0.0;
       const double ball_pos_x = wm.ball().pos().x;
       
       const double mark_pass_their_offence_line_weight = 0.0;
       const double mark_their_offence_line_weight = 0.0;
       const double guard_their_offence_line_weight = 0.0;
       const double positioning_their_offence_line_weight = 0.0;
       const double their_offence_line = wm.theirOffensePlayerLineX();
       
       const double mark_pass_dist_ball_from_me_weight = 0.0;
       const double mark_dist_ball_from_me_weight = 0.0;
       const double guard_dist_ball_from_me_weight = 0.0;
       const double positioning_dist_ball_from_me_weight = 0.0;
       const double dist_ball_from_me = wm.ball().pos().dist( wm.self().pos() );
       
       const double mark_pass_dist_ball_fom_target_opp_weight = 0.0;
       const double mark_dist_ball_fom_target_opp_weight = 0.0;
       const double guard_dist_ball_fom_target_opp_weight = 0.0;
       const double positioning_dist_ball_fom_target_opp_weight = 0.0;
       const double dist_ball_fom_target_opp = wm.ball().pos().dist( wm.theirPlayer( target_opp_unum )->pos() );
       
       const double mark_pass_dist_ball_from_goal_weight = 0.0;
       const double mark_dist_ball_from_goal_weight = 0.0;
       const double guard_dist_ball_from_goal_weight = 0.0;
       const double positioning_dist_ball_from_goal_weight = 0.0;
       const double dist_ball_from_goal = wm.ball().pos().dist( ServerParam::i().ourTeamGoalPos() );
       
       vector<MarkRoute> mark_actions;
       
       mark_actions.push_back( markPass   ( agent ) );
       mark_actions.push_back( mark       ( agent ) );
       
       int i = 0;
       
       double score = 0.0;
       double best_score = -1000000.0;
       int best_action = -1;
       
       for( i = 0  ; i < mark_actions.size() ; i++ )
       {
	 switch ( mark_actions.at( i ).mark_type_ )
	 {
	   case MARK_PASS :
	     score += teammate_six_x           * mark_pass_teammate_six_x_weight
	     +  ball_pos_x               * mark_pass_ball_pos_x_weight
	     +  their_offence_line       * mark_pass_their_offence_line_weight
	     +  dist_ball_from_me        * mark_pass_dist_ball_from_me_weight
	     +  dist_ball_fom_target_opp * mark_pass_dist_ball_fom_target_opp_weight
	     +  dist_ball_from_goal      * mark_dist_ball_from_goal_weight;
	     break;
	     
	   case MARK :
	     score += teammate_six_x           * mark_teammate_six_x_weight
	     +  ball_pos_x               * mark_ball_pos_x_weight
	     +  their_offence_line       * mark_their_offence_line_weight
	     +  dist_ball_from_me        * mark_dist_ball_from_me_weight
	     +  dist_ball_fom_target_opp * mark_dist_ball_fom_target_opp_weight
	     +  dist_ball_from_goal      * mark_dist_ball_from_goal_weight;
	     break;
	    
	     
	   case POSITIONING :
	     score += teammate_six_x           * positioning_teammate_six_x_weight
	     +  ball_pos_x               * positioning_ball_pos_x_weight
	     +  their_offence_line       * positioning_their_offence_line_weight
	     +  dist_ball_from_me        * positioning_dist_ball_from_me_weight
	     +  dist_ball_fom_target_opp * positioning_dist_ball_fom_target_opp_weight
	     +  dist_ball_from_goal      * positioning_dist_ball_from_goal_weight;
	     break;
	 }
	 
	 if( best_score <= score )
	 {
	   best_score = score;
	   best_action = i;
	 }
       }
       
       if( best_action == -1 )
       {
	 MarkRoute invalid_mark;
	 invalid_mark.dash_power = -10000;
	 invalid_mark.mark_type_ = NONE;
	 invalid_mark.target_pos_ = Vector2D::INVALIDATED;
	 return invalid_mark;
       }
       
       return mark_actions.at( best_action );
     }
