/*!
    <Bhv Pass Class.>
    Copyright (C) 2011  <Hessamoddin Hediehloo> <olampiad1370@yahoo.com> , <Mohammad Simchi> <m.simchi@yahoo.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#define DEBUG;

#include "bhv_pass.h"
#include "pass_object.h"

#include <rcsc/action/body_smart_kick.h>
#include <rcsc/player/player_object.h>
#include <rcsc/geom/vector_2d.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/common/server_param.h>



PassObject  Bhv_Pass::M_cached_pass;
std::vector< PassObject > Bhv_Pass::cached_pass_objects;

///*************************************************************************************/
bool Bhv_Pass::execute( PlayerAgent * agent )
{
  Vector2D pass_target = Vector2D::INVALIDATED;
  double first_speed = 0.0;
  
  if( virtualPassExecute( agent , pass_target , first_speed ) && pass_target.isValid() )
    return passExecute( agent , pass_target , first_speed);
  
  return false;
}


///*************************************************************************************/
bool Bhv_Pass::passExecute(PlayerAgent* agent , Vector2D pass_target , double first_speed)
{
  if( !pass_target.isValid() )
    return false;
  
  const double first_speed_thr = 0.8;
    
  const int max_step = ( (agent->world().gameMode().type() != GameMode::PlayOn && 
                          agent->world().gameMode().type() != GameMode::GoalKick_) ? 1 : 3 );
  
  if ( ! Body_SmartKick( pass_target,
			 first_speed,
			 first_speed * 0.96,
			 max_step ).execute( agent ) )
  {
    if ( agent->world().gameMode().type() != GameMode::PlayOn && 
	 agent->world().gameMode().type() != GameMode::GoalKick_ )
    {
      first_speed = std::min( agent->world().self().kickRate() * ServerParam::i().maxPower() , first_speed );
      Body_KickOneStep( pass_target,
			first_speed ).execute( agent );
      
      dlog.addText( Logger::ACTION, __FILE__": execute() one step kick" );
    }
    
    else
    {
      dlog.addText( Logger::ACTION , __FILE__": execute() failed to pass kick." );
      return false;
      
    }
  }
  /*! ------------------------------Say------------------------------ !*/

  if ( agent->config().useCommunication() && M_cached_pass.M_Receiver->unum() != Unum_Unknown ) 
  {
    Vector2D target_buf = pass_target - agent->world().ball().pos();
    target_buf.setLength( 0.5 );

    agent->addSayMessage( new PassMessage( M_cached_pass.M_Receiver->unum(),
					   pass_target + target_buf,
					   agent->effector().queuedNextBallPos(),
					   agent->effector().queuedNextBallVel() ) );
  }
  
  return true;
}

///*************************************************************************************/
bool Bhv_Pass::virtualPassExecute(PlayerAgent* agent , Vector2D & pass_target , double & first_speed)
{
  PassObject pass = getBestPass( agent );
  
  if( pass.M_Position.isValid() && pass.M_Score > -1000.0 )
  {
    pass_target = pass.M_Position;
    first_speed = pass.M_First_Speed;//getFirstSpeed(agent , pass_target);
    return true;
  }
  
  return false;
}

///*************************************************************************************/
PassObject Bhv_Pass::getBestPass(PlayerAgent* agent)
{
  const WorldModel &wm = agent->world();
  
  clear();
  
  PassObject pass_tmp;

  Vector2D target = Vector2D::INVALIDATED;
  const PlayerPtrCont & teammmate = wm.teammatesFromSelf();
  const PlayerPtrCont::const_iterator t_end_mates = teammmate.end();
  
  for ( PlayerPtrCont::const_iterator it = teammmate.begin() ; it != t_end_mates ; ++it )
  {
    if( (*it)->pos().x > wm.offsideLineX() )
      continue;
    
    if( (*it)->isSelf() || (*it)->goalie() || !(*it)->pos().isValid() )
      continue;
    
    if( wm.self().pos().x > -20.0 && ServerParam::i().ourPenaltyArea().contains( (*it)->pos() ) )
      continue;
    
    if( (*it)->pos().x < -20.0 )
      defPassPos( agent , *it );
    
    else if( (*it)->pos().x < 20.0 )
      midPassPos( agent , *it );
    
    else
      attPassPos( agent , *it );
    

  }

  
  if( cached_pass_objects.size() < 1 )
  {
    pass_tmp.M_Position = Vector2D::INVALIDATED;
    return pass_tmp;
  }

  int n = 0;
  double max_score = -1000000.0;
  Vector2D best_pass = Vector2D::INVALIDATED;
  std::vector<PassObject>::iterator it_end = cached_pass_objects.end();
  for ( std::vector<PassObject>::iterator it = cached_pass_objects.begin() ; it != it_end ; it++, n++ ) 
      {
	if ((*it).M_Score > max_score) 
	{
	  best_pass = (*it).M_Position;
	  max_score = (*it).M_Score;
	  pass_tmp  = (*it);
	}
	 
      }

  M_cached_pass = pass_tmp;
  return pass_tmp;
}

///*************************************************************************************/
void
Bhv_Pass::defPassPos( PlayerAgent* agent , PlayerObject * receiver )
{
  const WorldModel &wm = agent->world();
  Vector2D target = Vector2D::INVALIDATED;
  const PlayerPtrCont & teammmate = wm.teammatesFromSelf();
  const PlayerPtrCont::const_iterator t_end_mates = teammmate.end();
  PassObject pass_tmp;
    double dist = 0 , ang = 0 ;
    for ( dist = 0 ; dist < 3.1; dist += 0.75 )
      for ( ang = -180 ; ang < 180 ; ang +=20 )
      {
	if( wm.gameMode().type() == GameMode::GoalKick_ && 
	  ( wm.self().unum() == 2 || wm.self().unum() == 3 ) &&
	  ( receiver->unum() == 2 || receiver->unum() == 3 ) )
	  continue;
	  
	if( wm.gameMode().type() == GameMode::PlayOn && 
	    wm.self().pos().x <= -33.0 && wm.self().pos().absY() >= 25.0 && 
	    receiver->pos().absY() < 12.0 && receiver->pos().x < -40.0 )
	  continue;
	
	if( wm.self().unum() > 3 && wm.self().unum() < 7 && receiver->unum() > 1 && receiver->unum() < 4 )
	  continue;
	if ( dist == 0 ) { continue ; }
	target = receiver->pos() + Vector2D::polar2vector( dist , ang );
	if( target.dist( wm.self().pos() ) <= 4.0 || receiver->pos().dist( wm.self().pos() ) <= 4.0 )
	  continue;
	pass_tmp = PassObject( agent , target , receiver , getFirstSpeed(agent,target) );
	cached_pass_objects.push_back( pass_tmp );  
      }  
//   }
}

///**************************************************************************************/
void
Bhv_Pass::midPassPos(PlayerAgent* agent , PlayerObject * receiver )
{
  const WorldModel &wm = agent->world();
  Vector2D target;
  const PlayerPtrCont & teammmate = wm.teammatesFromSelf();
  const PlayerPtrCont::const_iterator t_end_mates = teammmate.end();
  PassObject pass_tmp;
   
      double dist  ,  ang  ;
      

      if  ( receiver->pos().y > 15 )
      {
	for ( dist = 0 ; dist < 10.1 ; dist += 3.0 )
	  for ( ang = 0 ; ( dist > 2.0 )? ang > -35 : ang > -180  ; (dist < 15 )? ang -=10 : ang-=5 )
	  {
	    if( wm.self().unum() > 3 && wm.self().unum() < 7 && receiver->unum() > 1 && receiver->unum() < 4 )
	      continue;
	    if (  dist > 3 && dist  < 15 ){continue;}
	    target = receiver->pos() + Vector2D::polar2vector( dist , ang );
	    if( target.dist( wm.self().pos() ) <= 4.0 || receiver->pos().dist( wm.self().pos() ) <= 4.0 )
	      continue;
	    pass_tmp = PassObject( agent , target  , receiver   , getFirstSpeed(agent,target));
	    cached_pass_objects.push_back( pass_tmp );  
	  }
      }

      else if  ( receiver->pos().y <= -15 )
      {
	for ( dist = 0 ; dist < 10.1 ; dist += 3.0 )
	  for ( ang = 0 ; ( dist > 2.0 )? ang < 35 : ang < 180  ; (dist < 15 )? ang +=10 : ang+=5 )
	  {
	    if( wm.self().unum() > 3 && wm.self().unum() < 7 && receiver->unum() > 1 && receiver->unum() < 4 )
	      continue;
	    if (  dist > 3 && dist  < 15 ){continue;}
	    target = receiver->pos() + Vector2D::polar2vector( dist , ang );
	    if( target.dist( wm.self().pos() ) <= 4.0 || receiver->pos().dist( wm.self().pos() ) <= 4.0 )
	      continue;
	    pass_tmp = PassObject( agent , target , receiver  , getFirstSpeed(agent,target) );
	    cached_pass_objects.push_back( pass_tmp );  
	  }
      }
      
      ////////////////////////////////////////////////////////////////////////////////////////////
      else
      {
	for ( dist = 0 ; dist < 11.1 ; dist += 3.0 )
	  for ( ang = -30 ; ang < 30 ; (dist < 10 )? ang +=10 : ang+=5 )
	  {
	    if( wm.self().unum() > 3 && wm.self().unum() < 7 && receiver->unum() > 1 && receiver->unum() < 4 )
	      continue;
	    if (  dist > 3 && dist  < 15 ){continue;}
	    target = receiver->pos() + Vector2D::polar2vector ( dist , ang );
	    if( target.dist( wm.self().pos() ) <= 4.0 || receiver->pos().dist( wm.self().pos() ) <= 4.0 )
	      continue;
	    pass_tmp = PassObject( agent , target , receiver ,getFirstSpeed(agent,target) );
	    cached_pass_objects.push_back( pass_tmp );  
	  }
	  ///******************************************************************/
	for ( dist = 0 ; dist < 1 ; dist += 0.25 )
	{
	  for ( ang = -30 ; ang > -160  ;  ang -=10 )
	  {
	    if( wm.self().unum() > 3 && wm.self().unum() < 7 && receiver->unum() > 1 && receiver->unum() < 4 )
	      continue;  
	    target = receiver->pos() + Vector2D::polar2vector ( dist , ang );
	    if( target.dist( wm.self().pos() ) <= 4.0 || receiver->pos().dist( wm.self().pos() ) <= 4.0 )
	      continue;
	    pass_tmp = PassObject( agent , target  , receiver , getFirstSpeed(agent,target));
	    cached_pass_objects.push_back( pass_tmp );  
	  }  
	  for ( ang = 30 ; ang < 60  ;  ang +=10 )
	  {
	    if( wm.self().unum() > 3 && wm.self().unum() < 7 && receiver->unum() > 1 && receiver->unum() < 4 )
	      continue;
	    target = receiver->pos() + Vector2D::polar2vector ( dist , ang );
	    if( target.dist( wm.self().pos() ) <= 4.0 || receiver->pos().dist( wm.self().pos() ) <= 4.0 )
	      continue;
	    pass_tmp = PassObject( agent , target , receiver  ,getFirstSpeed(agent,target));
	    cached_pass_objects.push_back( pass_tmp );  
	  } 
	}
      }
      ////////////////////////////////////////////////////////////////////////////////////////////
//     }
}

///**************************************************************************************/
void
Bhv_Pass::attPassPos(PlayerAgent* agent , PlayerObject * receiver)
{
  const WorldModel &wm = agent->world();
  Vector2D target;
  const PlayerPtrCont & teammmate = wm.teammatesFromSelf();
  const PlayerPtrCont::const_iterator t_end_mates = teammmate.end();
  PassObject pass_tmp;
    
    double dist  ,  ang ;
    for ( dist = 0 ; dist < 3.01 ; dist += 0.5 )
      for ( ang = -80 ; ang < 80 ;  ang += 10 )
      {
	if( wm.self().unum() > 1 && wm.self().unum() < 7 && receiver->unum() > 1 && receiver->unum() < 4 )
	  continue;
	target = receiver->pos() + Vector2D::polar2vector ( dist , ang );
	if( (target.dist( wm.self().pos() ) <= 4.0 || receiver->pos().dist( wm.self().pos() ) <= 4.0) && 
	    !ServerParam::i().theirPenaltyArea().contains( target ) )
	continue;
	pass_tmp = PassObject( agent , target , receiver , getFirstSpeed(agent,target));
	cached_pass_objects.push_back( pass_tmp );  
      }
for ( dist = 0 ; dist < 1.0; dist += 0.25 )
      for ( ang = 110 ; ang <= 180 ; ang += 10 )
      {
	if( wm.self().unum() > 1 && wm.self().unum() < 7 && receiver->unum() > 1 && receiver->unum() < 4 )
	  continue;
	target = receiver->pos() + Vector2D::polar2vector ( dist , ang );
	if( (target.dist( wm.self().pos() ) <= 4.0 || receiver->pos().dist( wm.self().pos() ) <= 4.0) && 
	    !ServerParam::i().theirPenaltyArea().contains( target ) )
	continue;
	pass_tmp = PassObject( agent , target , receiver , getFirstSpeed(agent,target));
	cached_pass_objects.push_back( pass_tmp );  
      }
for ( dist = 0 ; dist < 1.0 ; dist += 0.25 )
      for ( ang = -110 ; ang <= -180 ; ang -= 10 )
      {
	if( wm.self().unum() > 1 && wm.self().unum() < 7 && receiver->unum() > 1 && receiver->unum() < 4 )
	  continue;
	target = receiver->pos() + Vector2D::polar2vector ( dist , ang );
	if( (target.dist( wm.self().pos() ) <= 4.0 || receiver->pos().dist( wm.self().pos() ) <= 4.0) && 
	    !ServerParam::i().theirPenaltyArea().contains( target ) )
	continue;
	pass_tmp = PassObject( agent , target , receiver , getFirstSpeed(agent,target));
	cached_pass_objects.push_back( pass_tmp );  
      }      
//   }
}

///**************************************************************************************/

double Bhv_Pass::getFirstSpeed( PlayerAgent* agent , Vector2D target  )
{
 const WorldModel &wm = agent->world();
  double dist = wm.ball().pos().dist( target ) ;
     double end_speed = 1.5;
    double first_speed = 100.0;
    double first_speed_the = ServerParam::i().ballSpeedMax();
    
    do
    {
        first_speed
            = calc_first_term_geom_series_last
            ( end_speed,
              dist,
              ServerParam::i().ballDecay() );
        if ( first_speed < first_speed_the )
        {
            break;
        }
        end_speed -= 0.1;
    }
    while ( end_speed > 1.2 );
    return first_speed;
}

///**************************************************************************************/
void Bhv_Pass::clear()
{
  cached_pass_objects.clear();
  M_cached_pass.M_Passer = NULL;
  M_cached_pass.M_Position = Vector2D::INVALIDATED;
  M_cached_pass.M_Receiver = NULL;
  M_cached_pass.M_Score = -10000.0;
}
