/*!
    <Pass Object Class.>
    Copyright (C) 2012 <Hessamoddin Hediehloo> <olampiad1370@yahoo.com> , <Mohammad Simchi> <m.simchi@yahoo.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pass_object.h"

#include <rcsc/math_util.h>
#include <rcsc/time/timer.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/geom/line_2d.h>
#include <rcsc/player/player_agent.h>

using namespace rcsc; 

///**********************************************************************/
PassObject::PassObject()
{
  M_Agent     = NULL;
  M_Position  = Vector2D::INVALIDATED;
  M_Receiver  = NULL;
  M_Passer    = NULL;
}


///**********************************************************************/
PassObject::PassObject( PlayerAgent * agent , 
			Vector2D position , 
			PlayerObject * receiver ,
			double first_speed )
{
  M_Agent       = agent;
  M_Position    = position;
  M_Receiver    = receiver;
  M_First_Speed = first_speed;
  
  compute();  
}

///**********************************************************************/
void PassObject::compute()
{  
  const WorldModel & wm = M_Agent->world();
  
  const double dist_to_goal_weight      = 1.0;
  const double nearest_opp_dist_weight  = 1.0;
  const double receiver_dist_weight     = 1.0;
  
  double safety_ = safety();
  
  if( safety_ < 0.0 )
  {
    M_Score = -10000.0;
    return;
  }
  
  M_Score = M_Position.dist( ServerParam::i().theirTeamGoalPos() ) * dist_to_goal_weight + 
	    wm.getDistOpponentNearestTo( M_Position , 3 ) * nearest_opp_dist_weight +
	    M_Position.dist( M_Receiver->pos() ) * receiver_dist_weight;
}

///**********************************************************************/
double PassObject::safety()
{
  const WorldModel & wm = M_Agent->world();

   const PlayerPtrCont & opps = wm.opponentsFromSelf();   
  const PlayerPtrCont::const_iterator t_end_opps = opps.end();
  for ( PlayerPtrCont::const_iterator it = opps.begin() ; it != t_end_opps ; ++it ) 
  {
    if( opponentsCanReachToBall( *it ) || !teammatessCanReachToBall( M_Receiver ) )
    {
      return 1000.0;
    }    
  }
  return 1000.0;
}

///*************************************************************************/
int PassObject::cyclesObjectToReachPos(PlayerObject* o , Vector2D pos)
{ 
    double object_vel = (*o).vel().r();
    
    Vector2D object_pos = (*o).pos();
    double dist = object_pos.dist( pos );		
    return ( ceil( dist / object_vel ) );  
}

///**********************************************************************************/
bool PassObject::opponentsCanReachToBall( PlayerObject* opp )
{  
  const WorldModel & wm = M_Agent->world();
  
  Vector2D ball_pos = wm.ball().pos();
  
  const Vector2D ball_vell = wm.ball().vel();
  
  int i = 0;
  
  double term = 0.0;
  
  double first_speed = 0.0;
  const Vector2D first_ball_pos = ball_pos + ball_vell;
  const Vector2D required_ball_pos = M_Position - ( first_ball_pos );
  Vector2D first_ball_vel = required_ball_pos.setLengthVector(first_speed);  
  
  for( i = 0 ; ball_pos.dist( wm.ball().pos() ) < M_Position.dist( wm.ball().pos() ) ; i++ )
  {
    term = ( 1.0 - std::pow( ServerParam::i().ballDecay(), i ) ) / ( 1.0 - ServerParam::i().ballDecay() );
    
    ball_pos = first_ball_pos + (first_ball_vel * term);
       
    if( cyclesObjectToReachPos( opp , ball_pos ) <= i )
    {
      return false;
    }
  }

  return false;
}

///**********************************************************************************/
bool PassObject::teammatessCanReachToBall( PlayerObject* team )
{  
  const WorldModel & wm = M_Agent->world();
  
  Vector2D ball_pos = wm.ball().pos();
  
  const Vector2D ball_vell = wm.ball().vel();
  
  double term = 0.0;
  
  const Vector2D first_ball_pos = ball_pos + ball_vell;
  
  const Vector2D required_ball_pos = M_Position - ( first_ball_pos );
  
  const Vector2D first_ball_vel = required_ball_pos.setLengthVector(M_First_Speed);
  
  for( int i = 0 ; ball_pos.dist( wm.ball().pos() ) < M_Position.dist( wm.ball().pos() ) ; i++ )
  {
    term = ( 1.0 - std::pow( ServerParam::i().ballDecay(), i ) ) / ( 1.0 - ServerParam::i().ballDecay() );
    
    ball_pos = first_ball_pos + (first_ball_vel * term);    
    
    if( cyclesObjectToReachPos( team , M_Position ) <= i )
      return false;
  }
  
  return false;
}