// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_goalie.h"

#include "bhv_goalie_basic_move.h"
#include "bhv_goalie_chase_ball.h"
#include "bhv_goalie_free_kick.h"
#include "bhv_mark.h"
#include "ClearBall.h"
#include "rcsc/action/body_clear_ball.h"
#include "bhv_pass.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/body_hold_ball.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/world_model.h>
#include <rcsc/player/say_message_builder.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/rect_2d.h>

using namespace rcsc;

const std::string RoleGoalie::NAME( "Goalie" );
bool RoleGoalie :: ball_our_possession ;
int  RoleGoalie :: cycle_ball_poss ;
bool RoleGoalie::back_pass_risk ;

/*-------------------------------------------------------------------*/
/*!

 */
namespace {
rcss::RegHolder role = SoccerRole::creators().autoReg( &RoleGoalie::create,
                       RoleGoalie::name() );
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
RoleGoalie::execute( PlayerAgent * agent )
{
    static const Rect2D our_penalty( Vector2D( -ServerParam::i().pitchHalfLength(),
                                     -ServerParam::i().penaltyAreaHalfWidth() + 1.0 ),
                                     Size2D( ServerParam::i().penaltyAreaLength() - 1.0,
                                             ServerParam::i().penaltyAreaWidth() - 2.0 ) );

    


    // catchable
    if ( agent->world().time().cycle()
            > agent->world().self().catchTime().cycle() + ServerParam::i().catchBanCycle()
            && agent->world().ball().distFromSelf() < ServerParam::i().catchableArea() - 0.05
            && our_penalty.contains( agent->world().ball().pos() )  && !back_pass_risk )
    {
        dlog.addText( Logger::ROLE,
                      __FILE__": catchable. ball dist=%.1f, my_catchable=%.1f",
                      agent->world().ball().distFromSelf(),
                      ServerParam::i().catchableArea() );
        agent->doCatch();
        agent->setNeckAction( new Neck_TurnToBall() );
    }
    else if ( agent->world().self().isKickable() )
    {
        doKick( agent );
    }
    else
    {
        doMove( agent );
    }


    return true;
}

/*-------------------------------------------------------------------*/
/*!

 */
void
RoleGoalie::doKick( PlayerAgent * agent )
{
    if( !Bhv_Pass().execute( agent ) )
        if( ( agent->world().opponentsFromSelf().front()->pos().dist( agent->world().self().pos() ) <= 7 ) ||
                !Body_HoldBall().execute( agent ) || (agent->world().opponentsFromSelf().front()->posCount()>2) )
            Body_ClearBall().execute( agent );

    agent->setNeckAction( new Neck_ScanField() );
}

/*-------------------------------------------------------------------*/
/*!

 */
void
RoleGoalie::doMove( PlayerAgent * agent )
{
    const WorldModel & wm = agent->world() ;
    const PlayerPtrCont & teammate = wm.teammatesFromSelf() ;
    const PlayerPtrCont & opponent = wm.opponentsFromSelf() ;

    if(!teammate.empty() && !opponent.empty())
    {
        int opp_unum = -1;
        bool is_blocker = false;
        Bhv_Mark().markExecute( agent , opp_unum , is_blocker );
    }

    if ( Bhv_GoalieChaseBall::is_ball_chase_situation( agent ) )
    {
        Bhv_GoalieChaseBall().execute( agent );
    }

    else
    {
        Bhv_GoalieBasicMove().execute( agent );
    }
}


